"""
 -----------------------------------------------------------------------------
 Copyright (c) 2009-2017, Shotgun Software Inc

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

  - Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.

  - Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

  - Neither the name of the Shotgun Software Inc nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

-----------------------------------------------------------------------------
"""


# ----------------------------------------------------------------------------
# Errors
class MockgunError(Exception):
    """
    Base for all Mockgun related API Errors.
    These are errors that relate to mockgun specifically, for example
    relating to mockups setup and initialization. For operational errors,
    mockgun raises ShotgunErrors just like the Shotgun API.
    """



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

# Standard library imports
import os
from collections import OrderedDict
import warnings

# Local imports
from . import hooks, paths, compat, repos
from .module import Module, ModuleSpec, module_header, sort_modules
from .resolver import (
    ResolveError,
    Resolver,
    Activator,
    Copier,
    Localizer,
)
from .vendor import appdirs, yaml


__all__ = [
    'activate',
    'deactivate',
    'clone',
    'create',
    'localize',
    'publish',
    'resolve',
    'set_home_path',
    'get_home_path',
    'get_home_modules_path',
    'get_cache_path',
    'get_user_path',
    'get_user_modules_path',
    'get_modules',
    'get_module_paths',
    'add_module_path',
    'get_active_modules',
    'add_active_module',
    'remove_active_module',
    'get_repos',
    'get_repo',
    'add_repo',
    'get_config_path',
    'read_config',
    'write_config',
    'update_repo',
    'remove_repo',
]
_registry = {
    'repos': OrderedDict(),
}
_active_modules = []
missing = object()


def resolve(requirements):
    '''Resolve a list of module requirements.'''

    resolver = Resolver(get_repos())
    return resolver.resolve(requirements)


def localize(requirements, to_repo='home', overwrite=False):
    '''Localize a list of requirements.'''

    to_repo = get_repo(to_repo)

    # Resolve modules
    resolver = Resolver(get_repos())
    module_specs = resolver.resolve(requirements)

    # Localize modules from remote repos
    localizer = Localizer(to_repo)
    modules = localizer.localize(module_specs, overwrite)

    return modules


def activate(requirements):
    '''Resolve and active a list of module requirements.

    Usage:
        >>> cpenv.activate('moduleA', 'moduleB')

    Arguments:
        requirements (List[str]): List of module requirements

    Returns:
        list of Module objects that have been activated
    '''

    # Resolve modules
    resolver = Resolver(get_repos())
    module_specs = resolver.resolve(requirements)

    # Activate modules
    activator = Activator()
    modules = activator.activate(module_specs)

    return modules


def deactivate():
    '''Deactivates an environment by restoring all env vars to a clean state
    stored prior to activating environments
    '''
    # TODO:
    # Probably need to store a clean environment prior to activate.
    # In practice it's uncommon to activate then deactivate in the same
    # python session.
    pass


def create(where, name, version, **kwargs):
    '''Create a new module.

    Arguments:
        where (str): Path to new module
        name (str): Name of module
        version (str): Version of module
        description (str): Optional description of module
        author (str): Optional author of module
        email (str): Optional email address of author
        requires (list): Optional modules that this module depends on
        environment (dict): Optional environment variables

    Returns:
        Module object
    '''

    # Setup configuration defaults
    config = OrderedDict([
        ('name', name),
        ('version', version),
        ('description', kwargs.get('description', '')),
        ('author', kwargs.get('author', '')),
        ('email', kwargs.get('email', '')),
        ('requires', kwargs.get('requires', [])),
        ('environment', kwargs.get('environment', {})),
    ])

    # Check if module already exists
    where = paths.normalize(where)
    if os.path.isdir(where):
        raise OSError('Module already exists at "%s"' % where)

    # Create a Module object - does not yet exist on disk
    module = Module(where, name, version)

    # Run global precreate hook
    # Allows users to inject data into a config prior to creating a new module
    hooks.run_global_hook('pre_create', module, config)

    # Create module folder structure
    paths.ensure_path_exists(where)
    paths.ensure_path_exists(where + '/hooks')

    data = module_header + yaml.dump(
        dict(config),
        default_flow_style=False,
        sort_keys=False,
    )
    with open(paths.normalize(where, 'module.yml'), 'w') as f:
        f.write(data)

    # Run global postcreate hook
    # Allows users to perform some action after a module is created
    hooks.run_global_hook('post_create', module)

    return module


def remove(module, from_repo=None):
    '''Remove a module.'''

    if isinstance(module, Module):
        return module.remove()

    if isinstance(module, ModuleSpec):
        return module.repo.remove(module)

    if from_repo is None:
        raise ValueError('from_repo is required when removing module by name.')

    if isinstance(from_repo, compat.string_types):
        from_repo = get_repo(from_repo)

    module_spec = from_repo.find(module)[0]
    return from_repo.remove(module_spec)


def clone(module, from_repo=None, where=None, overwrite=False):
    '''Clone a module for local development.

    A typical development workflow using clone and publish:
        1. clone a module
        2. make changes
        3. test changes
        4. increment version in module.yml
        5. publish a new version of your module
    '''

    if not isinstance(module, (Module, ModuleSpec)):
        if from_repo is None:
            resolver = Resolver(get_repos())
            module_spec = resolver.resolve([module])[0]
        else:
            from_repo = get_repo(from_repo)
            module_spec = from_repo.find(module)[0]

    module = module_spec.repo.download(
        module_spec,
        where=paths.normalize(where or '.', module_spec.real_name),
        overwrite=overwrite,
    )

    return module


def publish(module, to_repo='home', overwrite=False):
    '''Publish a module to the specified repository.'''

    to_repo = get_repo(to_repo)

    if isinstance(module, compat.string_types):
        resolver = Resolver(get_repos())
        module = resolver.resolve([module])[0]

    if isinstance(module, ModuleSpec):
        if not isinstance(module.repo, repos.LocalRepo):
            raise ValueError('Can only from modules in local repos.')
        else:
            module = Module(module.path)

    published = to_repo.upload(module, overwrite)
    return published


def copy(module, from_repo, to_repo, overwrite=False):
    '''Copy a module from one repo to another.'''

    from_repo = get_repo(from_repo)
    to_repo = get_repo(to_repo)

    # Resolve module
    resolver = Resolver([from_repo])
    module_spec = resolver.resolve([module])[0]

    copier = Copier(to_repo)
    copied = copier.copy([module_spec], overwrite)
    return copied


def get_active_modules():
    '''Returns a list of active :class:`Module` s'''

    return _active_modules


def add_active_module(module):
    '''Add a module to CPENV_ACTIVE_MODULES environment variable.

    Arguments:
        module (Module): Module to add to CPENV_ACTIVE_MODULES
    '''

    if module not in _active_modules:
        _active_modules.append(module)

    _active_modules.sort(key=lambda m: m.real_name)

    module_names = os.pathsep.join([m.real_name for m in _active_modules])
    os.environ['CPENV_ACTIVE_MODULES'] = str(module_names)


def remove_active_module(module):
    '''Remove a module from CPENV_ACTIVE_MODULES environment variable.

    Arguments:
        module (Module): Module to remove from CPENV_ACTIVE_MODULES
    '''

    if module in _active_modules:
        _active_modules.remove(module)

    module_names = os.pathsep.join([m.real_name for m in _active_modules])
    os.environ['CPENV_ACTIVE_MODULES'] = str(module_names)


def set_home_path(path):
    '''Convenient function used to set the CPENV_HOME environment variable.'''

    # Set new home path
    home = paths.normalize(path)
    os.environ['CPENV_HOME'] = home
    _init_home_path(home)

    # Add new LocalRepo
    update_repo(repos.LocalRepo('home', get_home_modules_path()))

    return home


def _init_home_path(home):
    home_modules = paths.normalize(home, 'modules')
    home_cache = paths.normalize(home, 'cache')

    paths.ensure_path_exists(home)
    paths.ensure_path_exists(home_modules)
    paths.ensure_path_exists(home_cache)


def get_home_path():
    '''Returns the cpenv home directory.

    Default home paths:
        win - C:/ProgramData/cpenv
        mac - /Library/Application Support/cpenv
        linux - /usr/local/share/cpenv OR /usr/share/cpenv
    '''

    home_default = appdirs.site_data_dir('cpenv', appauthor=False)
    home = paths.normalize(os.getenv('CPENV_HOME', home_default))
    return home


def get_home_modules_path():
    '''Return the modules directory within the cpenv home directory.

    Default home modules paths:
        win - C:/ProgramData/cpenv/modules
        mac - /Library/Application Support/cpenv/modules
        linux - /usr/local/share/cpenv OR /usr/share/cpenv/modules
    '''

    return paths.normalize(get_home_path(), 'modules')


def get_cache_path(*parts):
    '''Return the cpenv cache directory within the cpenv home directory.

    Default cache paths:
        win - C:/ProgramData/cpenv/cache
        mac - /Library/Application Support/cpenv/cache
        linux - /usr/local/share/cpenv OR /usr/share/cpenv/cache

    Arguments:
        *parts (str) - List of path parts to join with cache path
    '''

    return paths.normalize(get_home_path(), 'cache', *parts)


def _init_user_path(user):
    '''Initialize user path.'''

    user_modules = paths.normalize(user, 'modules')
    paths.ensure_path_exists(user)
    paths.ensure_path_exists(user_modules)


def get_user_path():
    '''Returns the cpenv user directory.

    Default user paths:
        win - C:/Users/<username>/AppData/Roaming/cpenv
        mac - ~/Library/Application Support/cpenv
        linux - ~/.local/share/cpenv
    '''

    user_default = appdirs.user_data_dir('cpenv', appauthor=False)
    user = paths.normalize(user_default)

    return user


def get_user_modules_path():
    '''Returns the modules directory within the cpenv user directory.

    Default user paths:
        win - C:/Users/<username>/AppData/Roaming/cpenv/modules
        mac - ~/Library/Application Support/cpenv/modules
        linux - ~/.local/share/cpenv/modules
    '''

    return paths.normalize(get_user_path(), 'modules')


def get_module_paths():
    '''Returns a list of paths used to lookup local modules.

    The list of lookup paths contains:
        1. use modules path
        2. home modules path
        3. paths in CPENV_MODULES environment variable
    '''

    module_paths = [paths.normalize(os.getcwd()), get_user_modules_path()]

    cpenv_home_modules = get_home_modules_path()
    if cpenv_home_modules not in module_paths:
        module_paths.append(cpenv_home_modules)

    cpenv_modules_path = os.environ.get('CPENV_MODULES', None)
    if cpenv_modules_path:
        for module_path in cpenv_modules_path.split(os.pathsep):
            if module_path not in module_paths:
                module_paths.append(paths.normalize(module_path))

    return module_paths


def add_module_path(path):
    '''Add an additional lookup path for local modules.'''

    path = paths.normalize(path)
    module_paths = []

    # Get existing module lookup paths
    cpenv_modules = os.environ.get('CPENV_MODULES', '').split(os.pathsep)
    for module_path in cpenv_modules:
        if module_path:
            module_paths.append(module_path)

    # Add new module lookup path
    if path not in module_paths:
        module_paths.append(path)
        add_repo(repos.LocalRepo(path, path))

    # Persist in CPENV_MODULES
    os.environ['CPENV_MODULES'] = os.pathsep.join(module_paths)

    return module_paths


def get_modules(*requirements):
    '''Returns a list of available modules.'''

    if requirements:
        resolver = Resolver(get_repos())
        return sort_modules(resolver.resolve(requirements))

    modules = []

    for repo in get_repos():
        modules.extend(repo.list())

    return sort_modules(list(modules))


def update_repo(repo):
    '''Update a registered repo.'''

    _registry['repos'].update({repo.name: repo})


def add_repo(repo, idx=None):
    '''Register a Repo.

    Provide an idx to insert the Repo rather than append.
    '''

    if repo.name not in _registry['repos']:
        if idx is not None:
            items = list(_registry['repos'].items())
            items.insert(idx, (repo.name, repo))
            _registry['repos'] = OrderedDict(items)
        else:
            _registry['repos'][repo.name] = repo


def remove_repo(repo):
    '''Unregister a Repo.'''

    _registry['repos'].pop(repo.name, None)


def get_repo(name, **query):
    '''Get a repo by specifying an attribute to lookup'''

    if isinstance(name, repos.Repo):
        return name

    query['name'] = name

    for repo in get_repos():
        if all([getattr(repo, k, False) == v for k, v in query.items()]):
            return repo


def get_repos():
    '''Get a list of all registered Repos.'''

    return list(_registry['repos'].values())


def get_config_path():
    return paths.normalize(get_home_path(), 'config.yml')


def read_config(key=None, default=missing):
    '''Read the whole config or a specific key from disk.

    Examples:
        # Read whole config
        config = read_config()

        # Read one key
        repos = read_config('repos', {})
    '''

    config_path = get_config_path()
    if not os.path.isfile(config_path):
        return {}

    with open(config_path, 'r') as f:
        config = yaml.safe_load(f.read()) or {}

    if not key:
        return config

    config_key = config.get(key, missing)
    if config_key is missing:
        if default is missing:
            raise KeyError('Config has no key: ' + key)
        else:
            return default

    return config_key


def write_config(*args):
    '''Write the whole config or a specific key to disk.

    Examples:
        # Write whole config
        write_config({'repos': {}})

        # Write one config key
        write_config('repos', {})
    '''
    if len(args) == 1:
        config = args[0]
    elif len(args) == 2:
        config = read_config()
        config[args[0]] = args[1]
    else:
        raise ValueError('Expected 1 or 2 arguments got %s' % len(args))

    config_path = get_config_path()
    with open(config_path, 'w') as f:
        f.write(yaml.dump(config))


def _init():
    '''Responsible for initially configuraing cpenv.'''

    _init_home_path(get_home_path())
    _init_user_path(get_user_path())

    # Register all LocalRepos
    for path in get_module_paths():
        if path == paths.normalize(os.getcwd()):
            name = 'cwd'
        elif path == get_home_modules_path():
            name = 'home'
        elif path == get_user_modules_path():
            name = 'user'
        else:
            name = path
        add_repo(repos.LocalRepo(name, path))

    # Register repos from config
    configured_repos = read_config('repos', {})
    for name, config in configured_repos.items():
        repo_type = config.pop('type')
        repo_cls = repos.registry[repo_type]
        try:
            add_repo(repo_cls(**config))
        except Exception as e:
            warnings.warn('Failed to create %s repo named %s\nError: %s' % (
                repo_type,
                config['name'],
                str(e),
            ))

    # Set _active_modules from CPENV_ACTIVE_MODULES
    unresolved = []
    resolver = Resolver(get_repos())
    active_modules = os.getenv('CPENV_ACTIVE_MODULES')
    if active_modules:
        for module in active_modules.split(os.pathsep):
            if module:
                try:
                    resolved = resolver.resolve([module])[0]
                    _active_modules.append(resolved)
                except ResolveError:
                    unresolved.append(module)

    if unresolved:
        warnings.warn(
            'Unable to resolve %s from $CPENV_ACTIVE_MODULES:' % unresolved
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

# Standard library imports
import os
import sys
from collections import namedtuple
from string import Template

# Local imports
from . import compat, mappings, paths, repos
from .versions import ParseError, Version, default_version, parse_version
from .hooks import HookFinder, get_global_hook_path
from .vendor import yaml


module_header = '''
# Variables
# $MODULE - path to this module
# $PLATFORM - platform name (win, mac, linux)
# $PYVER - python version (2.7, 3.6...)

# Wrap variables in brackets when they are nested within a string.
#     DO 'this${variable}isnested/' NOT 'this$variableisnested'

'''


ModuleSpec = namedtuple(
    'ModuleSpec',
    ['name', 'real_name', 'qual_name', 'version', 'path', 'repo'],
)


class Module(object):

    def __init__(self, path, name=None, version=None, repo=None):

        self.path = paths.normalize(path)
        if repo is None:
            self.repo = repos.LocalRepo('tmp', paths.parent(self.path))
        else:
            self.repo = repo

        # Create HookFinder for this module
        self.hook_path = self.relative_path('hooks')
        self.hook_finder = HookFinder(
            self.hook_path,
            get_global_hook_path(),
        )

        # Setup config
        self.config_path = self.relative_path('module.yml')
        self.config_vars = {
            'MODULE': self.path,
            'PLATFORM': compat.platform,
            'PYVER': sys.version[:3],
        }
        self._raw_config = None
        self._config = None
        self._environ = None

        # Determine name, version, qual_name, and real_name

        if name and version:

            # Use name and version if explicitly passed
            self.name = name
            self.version = parse_version(version)

        else:

            # Check config
            if self.exists:
                name = self.config.get('name', None)
                version = self.config.get('version', None)
                if version:
                    version = parse_version(version)

            # Last resort parse module path
            if name is None or version is None:
                name, version = parse_module_path(self.path)

            self.name = name
            self.version = version

        self.qual_name = self.name + '-' + self.version.string

        base_name = os.path.basename(self.path)
        if self.version.string != base_name:
            self.real_name = base_name
        else:
            self.real_name = self.qual_name

    def __eq__(self, other):
        if hasattr(other, 'path'):
            return self.path == other.path
        else:
            return self.path == other

    def __hash__(self):
        return hash((self.__class__.__name__, self.path))

    def __repr__(self):
        return '<{}>(path="{}", name="{}", version={})'.format(
            self.__class__.__name__,
            self.path,
            self.name,
            self.version,
        )

    def as_spec(self):
        return ModuleSpec(
            name=self.name,
            real_name=self.real_name,
            qual_name=self.qual_name,
            version=self.version,
            path=self.path,
            repo=self.repo,
        )

    @classmethod
    def from_spec(cls, module_spec):
        return cls(
            name=module_spec.name,
            version=module_spec.version.string,
            path=module_spec.path,
            repo=module_spec.repo,
        )

    def relative_path(self, *args):
        '''Get a path relative to this module'''

        return paths.normalize(self.path, *args)

    def run_hook(self, hook_name):
        '''Run a module hook by name, fallback to global hook location.'''

        hook = self.hook_finder(hook_name)
        if hook:
            return hook.run(self)

    def spec(self, **kwargs):
        '''Return a ModuleSpec object for this Module.'''

        return ModuleSpec(
            name=kwargs.get('name', self.name),
            real_name=kwargs.get('real_name', self.real_name),
            qual_name=kwargs.get('qual_name', self.qual_name),
            path=kwargs.get('path', self.path),
            version=kwargs.get('version', self.version),
            repo=kwargs.get('repo', None),
        )

    def activate(self):
        '''Add this module to active modules'''

        from . import api
        self.run_hook('pre_activate')
        api.add_active_module(self)
        self.run_hook('post_activate')

    def remove(self):
        '''Delete this module'''

        from . import api
        self.run_hook('pre_remove')
        paths.rmtree(self.path)
        api.remove_active_module(self)
        self.run_hook('post_remove')

    @property
    def icon(self):
        return self.relative_path('icon.png')

    def has_icon(self):
        return os.path.isfile(self.icon)

    @property
    def description(self):
        return self.config.get('description', '')

    @property
    def author(self):
        return self.config.get('author', '')

    @property
    def email(self):
        return self.config.get('email', '')

    @property
    def is_active(self):
        from . import api
        return self.real_name in api.get_active_modules()

    @property
    def exists(self):
        paths = (
            self.path,
            self.config_path,
        )
        return all([os.path.exists(path) for path in paths])

    @property
    def raw_config(self):
        if self._raw_config is None:
            self._raw_config = read_raw_config(self.config_path)

        return self._raw_config

    @property
    def config(self):
        if self._config is None:

            if not self.raw_config:
                self._config = {}
                return self._config

            self._config = read_config(
                self.config_path,
                self.config_vars,
                self._raw_config,
            )

        return self._config

    @property
    def environment(self):
        if self._environ is None:
            env = self.config.get('environment', {})
            additional = {
                'CPENV_ACTIVE_MODULES': [self.real_name],
            }
            env = mappings.join_dicts(additional, env)
            self._environ = mappings.preprocess_dict(env)

        return self._environ


def read_raw_config(module_file):
    ''''Read the raw text data of a module.yml file'''

    with open(module_file, 'r') as f:
        return f.read()


def read_config(module_file, config_vars=None, data=None):
    '''Read and formats a module.yml file'''

    if config_vars is None:
        config_vars = {
            'MODULE': paths.parent(module_file),
            'PLATFORM': compat.platform,
            'PYVER': sys.version[:3],
        }

    if data is None:
        with open(module_file, 'r') as f:
            data = f.read()

    return yaml.safe_load(Template(data).safe_substitute(config_vars))


def sort_modules(modules, reverse=False):
    '''Sort a list of Modules or ModuleSpecs by version.'''

    return sorted(
        modules,
        key=lambda m: (m.real_name, m.version),
        reverse=reverse
    )


def is_module(path):
    '''Returns True if path refers to a module'''

    return os.path.exists(paths.normalize(path, 'module.yml'))


def parse_module_path(path, default_version=default_version):
    '''Return name and version from a module's path.'''

    basename = os.path.basename(path)

    try:
        version = parse_version(basename)
    except ParseError:
        if callable(default_version):
            default = default_version()
        else:
            default = default_version
        return basename, default

    head = basename.replace(version.string, '')
    if head:
        name = head.rstrip('_v').rstrip('-v').rstrip('-_')
    else:
        name = os.path.basename(os.path.dirname(path))

    return name, version


def parse_module_requirement(requirement, default_version=None):
    '''Given a requirement, return a name and version.'''

    if '\\' in requirement or '/' in requirement:
        # Probably a system path - lets parse it.
        return parse_module_path(requirement, default_version=default_version)

    try:
        version = parse_version(requirement)
    except ParseError:
        if callable(default_version):
            default = default_version()
        else:
            default = default_version
        return requirement, default

    name = requirement
    head = requirement.replace(version.string, '')
    if head:
        name = head.rstrip('_v').rstrip('-v').rstrip('-_').rstrip('<>=!')

    return name, version


def is_exact_match(requirement, module_spec):
    '''Is the module_spec an exact match for the provided requirement?'''

    name, version = parse_module_requirement(requirement)
    return (
        module_spec.qual_name == requirement or
        module_spec.real_name == requirement or
        (
            version and module_spec.name == name and
            module_spec.version == version
        )
    )


def is_partial_match(requirement, module_spec):
    '''Is the module_spec a partial match for the provided requirement?'''

    name, version = parse_module_requirement(requirement)
    return module_spec.name == name


def best_match(requirement, module_specs):

    name, version = parse_module_requirement(
        requirement,
        Version(0, 0, 0, None, None, '*')
    )

    best_match = None
    for module_spec in module_specs:
        if is_exact_match(requirement, module_spec):
            return module_spec
        if version < module_spec.version:
            if not best_match:
                best_match = module_spec
                continue
            if best_match.version < module_spec.version:
                best_match = module_spec

    return best_match



from .. import six

# Define all here to keep linters happy.  It should be overwritten by the code
# below, but if in the future __all__ is not defined in httplib2 this will keep
# things from breaking.
__all__ = []

# Import the proper implementation into the module namespace depending on the
# current python version.  httplib2 supports python 2/3 by forking the code rather
# than with a single cross-compatible module. Rather than modify third party code,
# we'll just import the appropriate branch here.
if six.PY3:
    # Generate ssl_error_classes
    import ssl as __ssl
    ssl_error_classes = (__ssl.SSLError, __ssl.CertificateError)
    del __ssl

    # get the python3 fork of httplib2
    from . import python3 as __httplib2_compat


else:
    # Generate ssl_error_classes
    from .python2 import SSLHandshakeError as __SSLHandshakeError  # TODO: shouldn't rely on this. not public
    ssl_error_classes = (__SSLHandshakeError,)
    del __SSLHandshakeError

    # get the python2 fork of httplib2
    from . import python2 as __httplib2_compat

# Import all of the httplib2 module.  Note that we can't use a star import because
# we need to import *everything*, not just what exists in __all__.
for __name in dir(__httplib2_compat):
    globals()[__name] = getattr(__httplib2_compat, __name)
del __httplib2_compat
del __name

# Add ssl_error_classes to __all__
__all__.append("ssl_error_classes")



# -*- coding: utf-8 -*-

# Standard library imports
import re
from collections import namedtuple
from functools import total_ordering


__all__ = [
    'ParseError',
    'Version',
    'parse_version',
    'default_version',
]


# Strange nuke version pattern
nuke_version_pattern = (
    r'(?P<major>\d+)'
    r'\.(?P<minor>\d+)'
    r'(?:v)'
    r'(?P<patch>\d+)$'
)
# Modified regex from semver.org
semver_version_pattern = (
    r'(?:v)?'
    r'(?P<major>\d+)'
    r'\.(?P<minor>\d+)'
    r'\.(?P<patch>\d+)'
    r'(?:-(?P<prerelease>(?:\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)'
    r'(?:\.(?:\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?'
    r'(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$'
)
simple_version_pattern = r'(?:v)?(?P<version>(\d+\.?)+)$'
VersionBase = namedtuple(
    'Version',
    ['major', 'minor', 'patch', 'prerelease', 'buildmetadata', 'string']
)


@total_ordering
class Version(VersionBase):

    _defaults = {
        'major': 0,
        'minor': 0,
        'patch': 0,
        'prerelease': None,
        'buildmetadata': None
    }

    def __str__(self):
        return self.string

    def __hash__(self):
        return super(Version, self).__hash__()

    def _comparable(self):
        '''Generate a comparison key for a Version object.

        We use 'zzzzzzzz' as a default in lieu of complicated conditional logic
        to force releases to take precendence over prereleases.
        '''
        return (
            self.major,
            self.minor,
            self.patch,
            self.prerelease or 'zzzzzzzz',
            self.buildmetadata or 'zzzzzzzz',
        )

    def __lt__(self, other):
        if not isinstance(other, Version):
            raise ValueError('Can only compare two Version objects.')

        return self._comparable() < other._comparable()

    def __eq__(self, other):
        if not isinstance(other, Version):
            raise ValueError('Can only compare two Version objects.')

        return self._comparable() == other._comparable()


class ParseError(Exception):
    '''Raised when a parse method fails.'''


def parse_version(string):
    '''Parse and return a Version from the provided string.

    Supports:
      - semver / calver
      - simple versions like: 10, v2, 1.0, 2.2.4

    Arguments:
        string (str): String to parse version from.

    Returns:
        Version object

    Raises:
        ParseError when a version can not be parsed.
    '''
    # Parse weird nuke versioning
    match = re.search(nuke_version_pattern, string)
    if match:
        return Version(
            major=int(match.group('major')),
            minor=int(match.group('minor')),
            patch=int(match.group('patch')),
            prerelease=None,
            buildmetadata=None,
            string=match.group(0)
        )

    # Parse Semver / Calver
    match = re.search(semver_version_pattern, string)
    if match:
        return Version(
            major=int(match.group('major')),
            minor=int(match.group('minor')),
            patch=int(match.group('patch')),
            prerelease=match.group('prerelease'),
            buildmetadata=match.group('buildmetadata'),
            string=match.group(0),
        )

    # Parse Simple version
    match = re.search(simple_version_pattern, string)
    if match:
        kwargs = dict(Version._defaults)
        kwargs['string'] = match.group(0)
        version_parts = match.group('version').split('.')
        for part, part_name in zip(version_parts, ['major', 'minor', 'patch']):
            kwargs[part_name] = int(part)
        return Version(**kwargs)

    raise ParseError('Could not parse version from %s' % string)


def default_version():
    return Version(
        major=0,
        minor=1,
        patch=0,
        prerelease=None,
        buildmetadata=None,
        string='0.1.0',
    )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

# Standard library imports
import argparse
import os
import re
import sys
import atexit

# Third party imports
import tqdm

# Local imports
import cpenv
from cpenv import (
    Copier,
    LocalRepo,
    Localizer,
    Reporter,
    Resolver,
    ResolveError,
    api,
    cli,
    shell,
    paths,
)
from cpenv.module import (
    Module,
    best_match,
    is_partial_match,
    parse_module_path,
    sort_modules,
)


class CpenvCLI(cli.CLI):
    '''
    Create, activate and manage Modules.

    A Module is a folder containing a module.yml file describing the Module's
    name, version, requirements, and environment variables. Module's can also
    contain a hooks folder with python files that respond to specific events.
    '''

    name = 'cpenv'
    usage = 'cpenv [-h] <command> [<args>...]'

    def commands(self):
        return [
            Activate(self),
            Clone(self),
            Copy(self),
            Create(self),
            Info(self),
            Edit(self),
            List(self),
            Localize(self),
            Publish(self),
            Remove(self),
            Repo(self),
            Version(self),
        ]


class Info(cli.CLI):
    '''Show Module info'''

    usage = 'cpenv info [-h] [<modules>...]'

    def setup_parser(self, parser):
        parser.add_argument(
            'modules',
            help='Space separated list of modules.',
            nargs='+',
        )

    def run(self, args):
        cli.echo()
        try:
            module_specs = api.resolve(args.modules)
        except ResolveError:
            sys.exit(1)

        cli.echo()
        for spec in module_specs:
            cli.echo(cli.format_section(
                spec.real_name,
                [(k, str(v)) for k, v in sorted(spec._asdict().items())]
            ))
            cli.echo()


class Edit(cli.CLI):
    '''Open a module in a text editor.

    Editor Lookup:
        1. CPENV_EDITOR environment variable
        2. EDITOR environment variable
        3. subl (default editor)
    '''

    usage = 'cpenv edit [-h] <module>'

    def setup_parser(self, parser):
        parser.add_argument(
            'module',
            help='Module to clone.',
        )

    def run(self, args):
        cli.echo()
        try:
            module_spec = api.resolve([args.module])[0]
        except ResolveError:
            sys.exit(1)

        if not isinstance(module_spec.repo, LocalRepo):
            cli.echo('%s - %s' % (module_spec.qual_name, module_spec.path))
            cli.echo('Error: Can only edit modules in local repositories.')
            sys.exit(1)

        editor = os.getenv('CPENV_EDITOR', os.getenv('EDITOR', 'subl'))
        cli.echo('Opening %s in %s.' % (module_spec.path, editor))
        shell.run(editor, module_spec.path)


class Activate(cli.CLI):
    '''Activate a list of Modules.'''

    usage = 'cpenv activate [-h] [<modules>...]'

    def setup_parser(self, parser):
        parser.add_argument(
            'modules',
            help='Space separated list of modules.',
            nargs='+',
        )

    def run(self, args):
        cli.echo()
        try:
            api.activate(args.modules)
        except ResolveError:
            sys.exit(1)

        cli.echo('- Launching subshell...')
        cli.echo()
        cli.echo('  Type "exit" to deactivate.')
        cli.echo()
        shell.launch('[*]')


class Clone(cli.CLI):
    '''
    Clone a Module for development.

    The following repos are available by default:
        home - Local repo pointing to a computer-wide cpenv directory
        user - Local repo pointing to a user-specific cpenv directory

    For a full listing of available repos use the repo cli command:
        cpenv repo list
    '''

    def setup_parser(self, parser):
        parser.add_argument(
            'module',
            help='Module to clone.',
        )
        parser.add_argument(
            'where',
            help='Destination directory. (./<module_name>)',
            nargs='?',
            default=None,
        )
        parser.add_argument(
            '--from_repo',
            help='Specific repo to clone from.',
            default=None,
        )
        parser.add_argument(
            '--overwrite',
            help='Overwrite the destination directory. (False)',
            action='store_true',
        )

    def run(self, args):

        try:
            if not args.from_repo:
                module_spec = api.resolve([args.module])[0]
            else:
                from_repo = api.get_repo(args.from_repo)
                module_spec = from_repo.find(args.module)[0]
        except Exception:
            cli.echo()
            cli.echo('Error: Failed to resolve ' + args.module)
            sys.exit(1)

        where = paths.normalize(args.where or '.', module_spec.real_name)
        if os.path.isdir(where):
            cli.echo('Error: Directory already exists - ' + where)
            sys.exit(1)

        cli.echo('- Cloning %s...' % module_spec.real_name)
        cli.echo()
        try:
            module = module_spec.repo.download(
                module_spec,
                where=where,
                overwrite=args.overwrite,
            )
        except Exception as e:
            cli.echo()
            cli.echo('Error: ' + str(e))
            sys.exit(1)

        cli.echo()
        cli.echo('Navigate to the following folder to make changes:')
        cli.echo('  ' + module.path)
        cli.echo()
        cli.echo("Use one of the following commands to publish your changes:")
        cli.echo('  cpenv publish .')
        cli.echo('  cpenv publish . --to_repo="repo_name"')


class Create(cli.CLI):
    '''Create a new Module.'''

    def setup_parser(self, parser):
        parser.add_argument(
            'where',
            help='Path to new module',
        )

    def run(self, args):
        where = paths.normalize(args.where)
        if os.path.isdir(where):
            cli.echo()
            cli.echo('Error: Can not create module in existing directory.')
            sys.exit(1)

        default_name, default_version = parse_module_path(where)

        cli.echo()
        cli.echo('This command will guide you through creating a new module.')
        cli.echo()
        name = cli.prompt('  Module Name [%s]: ' % default_name)
        version = cli.prompt('  Version [%s]: ' % default_version.string)
        description = cli.prompt('  Description []: ')
        author = cli.prompt('  Author []: ')
        email = cli.prompt('  Email []: ')

        cli.echo()
        cli.echo('- Creating your new Module...', end='')
        module = api.create(
            where=where,
            name=name or default_name,
            version=version or default_version.string,
            description=description,
            author=author,
            email=email,
        )
        cli.echo('OK!')

        cli.echo()
        cli.echo('  ' + module.path)

        cli.echo()
        cli.echo('Steps you might take before publishing...')
        cli.echo()
        cli.echo('  - Include binaries your module depends on')
        cli.echo('  - Edit the module.yml file')
        cli.echo('    - Add variables to the environment section')
        cli.echo('    - Add other modules to the requires section')
        cli.echo('  - Add python hooks like post_activate')
        cli.echo()


class List(cli.CLI):
    '''List active and available Modules.'''

    def setup_parser(self, parser):
        parser.add_argument(
            'requirement',
            help='Space separated list of modules.',
            nargs='?',
            default=None,
        )
        parser.add_argument(
            '--verbose', '-v',
            help='Print more module info.',
            action='store_true',
        )

    def run(self, args):

        found_modules = False

        cli.echo()
        active_modules = api.get_active_modules()
        if args.requirement:
            active_modules = [
                m for m in active_modules
                if is_partial_match(args.requirement, m)
            ]

        if active_modules:
            found_modules = True
            cli.echo(cli.format_columns(
                '[*] Active',
                [m.real_name for m in sort_modules(active_modules)],
            ))
            cli.echo()

        for repo in api.get_repos():
            if args.requirement:
                repo_modules = repo.find(args.requirement)
            else:
                repo_modules = repo.list()

            module_names = []
            for module in sort_modules(repo_modules):
                if module in active_modules:
                    module_names.append('* ' + module.real_name)
                else:
                    module_names.append('  ' + module.real_name)

            if module_names:
                found_modules = True
                if repo.name != repo.path:
                    header = repo.name + ' - ' + repo.path
                else:
                    header = repo.name
                cli.echo(cli.format_columns(
                    '[ ] ' + header,
                    module_names,
                    indent='  ',
                ))
                cli.echo()

        if not found_modules:
            cli.echo('No modules are available.')


class Localize(cli.CLI):
    '''Localize a list of Modules.

    Downloads modules from a remote Repo and places them in the home LocalRepo
    by default. Use the --to_repo option to specify a LocalRepo.
    '''

    def setup_parser(self, parser):
        parser.add_argument(
            'modules',
            help='Space separated list of modules.',
            nargs='+',
        )
        parser.add_argument(
            '--to_repo', '-r',
            help='Specific repo to localize to. (first match)',
            default=None,
        )
        parser.add_argument(
            '--overwrite', '-o',
            help='Overwrite the destination directory. (False)',
            action='store_true',
        )

    def run(self, args):

        cli.echo()

        if args.to_repo:
            to_repo = api.get_repo(name=args.to_repo)
        else:
            repos = [r for r in api.get_repos() if isinstance(r, LocalRepo)]
            to_repo = prompt_for_repo(
                repos,
                'Choose a repo to localize to',
                default_repo_name='home',
            )

        # Resolve modules
        try:
            resolved = api.resolve(args.modules)
        except ResolveError:
            sys.exit(1)

        # Localize modules from remote repos
        remote_modules = [
            spec for spec in resolved
            if not isinstance(spec.repo, LocalRepo)
        ]

        if not len(remote_modules):
            cli.echo('All modules are already local...')
            sys.exit()

        try:
            localizer = Localizer(to_repo)
            localizer.localize(
                remote_modules,
                overwrite=args.overwrite,
            )
        except Exception as e:
            cli.echo('Error: ' + str(e))
            sys.exit(1)


class Copy(cli.CLI):
    '''Copy Modules from one Repo to another.'''

    def setup_parser(self, parser):
        parser.add_argument(
            'modules',
            help='Space separated list of modules.',
            nargs='+',
        )
        parser.add_argument(
            '--from_repo',
            help='Download from',
            default=None,
        )
        parser.add_argument(
            '--to_repo',
            help='Upload to',
            default=None,
        )
        parser.add_argument(
            '--overwrite',
            help='Overwrite the destination directory. (False)',
            action='store_true',
        )

    def run(self, args):

        cli.echo()

        if args.from_repo:
            from_repo = api.get_repo(args.from_repo)
        else:
            from_repo = prompt_for_repo(
                api.get_repos(),
                'Download from',
                default_repo_name='home',
            )

        if args.to_repo:
            to_repo = api.get_repo(args.to_repo)
        else:
            repos = api.get_repos()
            repos.remove(from_repo)
            to_repo = prompt_for_repo(
                repos,
                'Upload to',
                default_repo_name='home',
            )

        resolver = Resolver([from_repo])
        module_specs = resolver.resolve(args.modules)
        cli.echo()

        choice = cli.prompt('Copy these modules to %s?[y/n] ' % to_repo.name)
        if choice.lower() not in ['y', 'yes', 'yup']:
            cli.echo('Aborted.')
            sys.exit(1)
        cli.echo()

        copier = Copier(to_repo)
        copier.copy(module_specs, args.overwrite)


class Publish(cli.CLI):
    '''Publish a Module to a repo.'''

    def setup_parser(self, parser):
        parser.add_argument(
            'module',
            help='Path of module to publish. (".")',
            default='.',
            nargs='?',
        )
        parser.add_argument(
            '--to_repo',
            help='Specific repo to clone from.',
            default=None,
        )
        parser.add_argument(
            '--overwrite',
            help='Overwrite the destination directory. (False)',
            action='store_true',
        )

    def run(self, args):

        cli.echo()

        # Get repo
        if args.to_repo:
            to_repo = api.get_repo(name=args.to_repo)
        else:
            to_repo = prompt_for_repo(
                api.get_repos(),
                'Choose a repo to publish to',
                default_repo_name='home',
            )

        # Resolve module
        resolver = Resolver(api.get_repos())
        module_spec = resolver.resolve([args.module])[0]
        cli.echo()

        # Confirm publication
        choice = cli.prompt('Publish module to %s?[y/n] ' % to_repo.name)
        if choice.lower() not in ['y', 'yes', 'yup']:
            cli.echo('Aborted.')
            sys.exit(1)

        # Publish
        module = Module(module_spec.path)
        published = to_repo.upload(module, args.overwrite)
        cli.echo()

        cli.echo('Activate your module:')
        cli.echo('  cpenv activate %s' % published.real_name)
        cli.echo()


class Remove(cli.CLI):
    '''Permanently remove a Module from a Repo.'''

    def setup_parser(self, parser):
        parser.add_argument(
            'modules',
            help='Space separated list of modules.',
            nargs='+',
        )
        parser.add_argument(
            '--from_repo',
            help='Specific repo to remove from.',
            default=None,
        )
        parser.add_argument(
            '--quiet',
            help='Overwrite the destination directory. (False)',
            action='store_true',
        )

    def run(self, args):

        cli.echo()

        if args.from_repo:
            from_repo = api.get_repo(name=args.from_repo)
        else:
            from_repo = prompt_for_repo(
                api.get_repos(),
                'Choose a repo to remove module from',
                default_repo_name='home',
            )

        cli.echo('- Finding modules in %s...' % from_repo.name)
        cli.echo()
        modules_to_remove = []
        for requirement in args.modules:
            module = best_match(requirement, from_repo.find(requirement))
            if not module:
                cli.echo('Error: %s not found...' % requirement)
                sys.exit(1)
            cli.echo('  %s - %s' % (module.real_name, module.path))
            modules_to_remove.append(module)

        cli.echo()
        choice = cli.prompt('Delete these modules?[y/n] ')
        if choice.lower() not in ['y', 'yes', 'yup']:
            cli.echo('Aborted.')
            sys.exit(1)

        cli.echo()
        cli.echo('- Removing modules...')
        cli.echo()
        for module in modules_to_remove:
            cli.echo('  ' + module.real_name)
            api.remove(module, from_repo)

        cli.echo()
        cli.echo('Successfully removed modules.')
        cli.echo()


class Repo(cli.CLI):
    '''
    Manage, Configure and list Repos.

    Repos are sources of Modules. A Repo can be local or remote. Modules
    stored in remote Repos will be localized to a local repo prior to
    activation or explicitly by using "cpenv localize".
    '''

    def commands(self):
        return [
            ListRepo(self),
            AddRepo(self),
            RemoveRepo(self),
        ]


class ListRepo(cli.CLI):
    '''List configured repos'''

    name = 'list'

    def run(self, args):

        cli.echo()
        cli.echo('Repos in order of resolution:')
        for i, repo in enumerate(api.get_repos()):
            if repo.path == repo.name:
                cli.echo('  [{}] {}  {}'.format(
                    i,
                    type(repo).__name__,
                    repo.path,
                ))
            else:
                cli.echo('  [{}] {}  {}  {}'.format(
                    i,
                    type(repo).__name__,
                    repo.name,
                    repo.path,
                ))


class AddRepo(cli.CLI):
    '''Add a new repo.

    LocalRepo:
        cpenv repo add custom --path=~/custom_repo

    ShotgunRepo:
        cpenv repo add my_shotgun --type=shotgun --base_url=https://my.shotgunstudio.com --script_name=cpenv --api_key=secret
    '''

    name = 'add'

    def setup_parser(self, parser):
        parser.add_argument(
            '--type',
            help='Type of repo',
            choices=['local', 'shotgun'],
            default='local',
        )
        parser.add_argument(
            'name',
            help='Name of the repo',
        )
        parser.add_argument(
            'type_args',
            help='Type specific arguments.',
            nargs=argparse.REMAINDER,
        )

    def parse_type_args(self, type_args):
        pattern = (
            r'-{1,2}(?P<param>[a-zA-Z0-9_]+)=*\s*'
            r'"?(?P<value>.+)"?'
        )
        kwargs = {}
        for arg in type_args:
            match = re.search(pattern, arg)
            param = match.group('param')
            raw_value = '"%s"' % match.group('value').strip()
            value = cli.safe_eval(raw_value)
            kwargs[param] = value
        return kwargs

    def run(self, args):

        # Parse type_args or args that are specific to a given Repo type
        repo_kwargs = self.parse_type_args(args.type_args)
        repo_type = repo_kwargs.pop('type', args.type)

        cli.echo()
        if repo_type not in cpenv.repos.registry:
            cli.echo('Error: %s is not a registered repo type.' % args.type)
            cli.echo('Choose from: ' + ', '.join(cpenv.repos.registry.keys()))
            sys.exit(1)

        if cpenv.get_repo(args.name):
            cli.echo('Error: Repo named %s already exists.' % args.name)
            sys.exit(1)

        repo_cls = cpenv.repos.registry[repo_type]
        repo_kwargs['name'] = args.name
        cli.echo('- Checking %s args...' % repo_cls.__name__, end='')
        try:
            repo_cls(**repo_kwargs)
        except Exception as e:
            cli.echo('OOPS!')
            cli.echo()
            cli.echo('Error: Failed to initialize %s' % repo_cls.__name__)
            cli.echo(str(e))
            sys.exit(1)
        cli.echo('OK!')
        cli.echo()

        cli.echo('- Adding repo to config...', end='')
        repo_kwargs['type'] = repo_type
        repo_config = cpenv.read_config('repos', {})
        repo_config[args.name] = repo_kwargs
        cpenv.write_config('repos', repo_config)
        cli.echo('OK!')
        cli.echo()


class RemoveRepo(cli.CLI):
    '''Remove repo by name.'''

    name = 'remove'

    def setup_parser(self, parser):
        parser.add_argument(
            'name',
            help='Name of the repo',
        )

    def run(self, args):
        cli.echo()
        if args.name in ['home', 'user', 'cwd']:
            cli.echo('Error: Can not remove %s repo.' % args.name)

        repo_config = cpenv.read_config('repos', {})
        if args.name not in repo_config:
            cli.echo('Error: Repo named %s not found.' % args.name)
            sys.exit(1)

        cli.echo('- Removing repo from config...', end='')
        repo_config.pop(args.name)
        cpenv.write_config('repos', repo_config)
        cli.echo('OK!')
        cli.echo()


class Version(cli.CLI):
    '''Show version information.'''

    def run(self, args):

        cli.echo()
        cli.echo(cli.format_section(
            'Version Info',
            [
                ('name', cpenv.__name__),
                ('version', cpenv.__version__),
                ('url', cpenv.__url__),
                ('package', paths.parent(cpenv.__file__)),
                ('path', api.get_module_paths()),
            ]
        ))
        cli.echo()

        # List package versions
        dependencies = []
        try:
            import Qt
            dependencies.extend([
                ('Qt.py', Qt.__version__),
                ('Qt Binding', Qt.__binding__ + '-' + Qt.__binding_version__),
            ])
        except ImportError:
            pass

        if not dependencies:
            return

        cli.echo(cli.format_section('Dependencies', dependencies), end='\n\n')


def prompt_for_repo(repos, message, default_repo_name='home'):
    '''Prompt a user to select a repository'''

    default = 0
    for i, from_repo in enumerate(repos):
        if from_repo.name == default_repo_name:
            default = i
        if from_repo.name == from_repo.path:
            line = '  [{}] {}'.format(i, from_repo.path)
        else:
            line = '  [{}] {} - {}'.format(
                i,
                from_repo.name,
                from_repo.path,
            )
        cli.echo(line)

    # Prompt user to choose a repo defaults to home
    cli.echo()
    choice = cli.prompt('{} [{}]: '.format(message, default))

    if not choice:
        choice = default
        cli.echo()
        return repos[choice]

    try:
        choice = int(choice)
        value_error = False
    except ValueError:
        value_error = True

    if value_error or choice > len(repos) - 1:
        cli.echo()
        cli.echo('Error: {!r} is not a valid choice'.format(choice))
        sys.exit(1)

    cli.echo()

    # Get the repo the user chose
    return repos[choice]


class CliReporter(Reporter):

    def __init__(self):
        self._bars = {}

    def get_bar_style(self, desc, total, unit=None):
        return {
            'desc': desc,
            'total': total,
            'bar_format': '  {desc} {bar} {n_fmt}/{total_fmt}',
            'unit': unit or 'iB',
            'unit_scale': True,
        }

    def start_resolve(self, requirements):
        cli.echo('- Resolving requirements...')
        cli.echo()

    def resolve_requirement(self, requirement, module_spec):
        cli.echo('  %s - %s' % (module_spec.real_name, module_spec.path))

    def end_resolve(self, resolved, unresolved):
        cli.echo()
        if unresolved:
            cli.echo('Error: Failed to resolve %s' % ', '.join(unresolved))

    def start_localize(self, module_specs):
        for spec in module_specs:
            if not isinstance(spec.repo, LocalRepo):
                cli.echo('- Localizing modules...')
                cli.echo()
                return

    def end_localize(self, modules):
        cli.echo()

    def start_progress(self, label, max_size, data):

        if 'download' in label.lower():
            spec = data['module_spec']
            cli.echo(
                '  Downloading %s from %s...' %
                (spec.qual_name, spec.repo.name)
            )
            desc = spec.real_name
        elif 'upload' in label.lower():
            module = data['module']
            to_repo = data['to_repo']
            cli.echo(
                '  Uploading %s to %s...' %
                (module.qual_name, to_repo.name)
            )
            desc = module.real_name
        else:
            desc = label

        style = self.get_bar_style(desc, max_size, data.get('unit', None))
        self._bars[label] = tqdm.tqdm(**style)

    def update_progress(self, label, chunk_size, data):
        bar = self._bars.get(label, None)
        if bar:
            bar.update(chunk_size)

    def end_progress(self, label, data):
        bar = self._bars.pop(label, None)
        if bar:
            bar.close()


def perform_self_version_check():
    from . import _self_version_check
    is_latest, current, latest = _self_version_check.is_latest_version()
    if not is_latest:
        _self_version_check.warn_newer_version_available(current, latest)


def main():
    atexit.register(perform_self_version_check)
    cpenv.set_reporter(CliReporter)
    cli.run(CpenvCLI, sys.argv[1:])


if __name__ == '__main__':
    main()



#! /opt/local/bin/python

# Copyright (c) 2019 Shotgun Software Inc.
#
# CONFIDENTIAL AND PROPRIETARY
#
# This work is provided "AS IS" and subject to the Shotgun Pipeline Toolkit
# Source Code License included in this distribution package. See LICENSE.
# By accessing, using, copying or modifying this work you indicate your
# agreement to the Shotgun Pipeline Toolkit Source Code License. All rights
# not expressly granted therein are reserved by Shotgun Software Inc.

# ----------------------------------------------------------------------------
#  SG_TIMEZONE module
#  this is rolled into the this shotgun api file to avoid having to require
#  current users of api2 to install new modules and modify PYTHONPATH info.
# ----------------------------------------------------------------------------

from datetime import tzinfo, timedelta
import time as _time


class SgTimezone(object):
    '''
    Shotgun's server infrastructure is configured for Coordinated Universal
    Time (UTC). In order to provide relevant local timestamps to users, we wrap
    the datetime module's tzinfo to provide convenient conversion methods.
    '''

    ZERO = timedelta(0)
    STDOFFSET = timedelta(seconds=-_time.timezone)
    if _time.daylight:
        DSTOFFSET = timedelta(seconds=-_time.altzone)
    else:
        DSTOFFSET = STDOFFSET
    DSTDIFF = DSTOFFSET - STDOFFSET

    def __init__(self):
        self.utc = UTC()
        self.local = LocalTimezone()

    @classmethod
    def UTC(cls):
        '''
        For backwards compatibility, from when UTC was a nested class,
        we allow instantiation via SgTimezone
        '''
        return UTC()

    @classmethod
    def LocalTimezone(cls):
        '''
        For backwards compatibility, from when LocalTimezone was a nested
        class, we allow instantiation via SgTimezone
        '''
        return LocalTimezone()


class UTC(tzinfo):
    '''
    Implementation of datetime's tzinfo to provide consistent calculated
    offsets against Coordinated Universal Time (UTC)
    '''

    def utcoffset(self, dt):
        return SgTimezone.ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return SgTimezone.ZERO


class LocalTimezone(tzinfo):
    '''
    Implementation of datetime's tzinfo to provide convenient conversion
    between Shotgun server time and local user time
    '''

    def utcoffset(self, dt):
        '''
        Difference between the user's local timezone and UTC timezone in seconds
        '''
        if self._isdst(dt):
            return SgTimezone.DSTOFFSET
        else:
            return SgTimezone.STDOFFSET

    def dst(self, dt):
        '''
        Daylight savings time (dst) offset in seconds
        '''
        if self._isdst(dt):
            return SgTimezone.DSTDIFF
        else:
            return SgTimezone.ZERO

    def tzname(self, dt):
        '''
        Name of the user's local timezone, including a reference
        to daylight savings time (dst) if applicable
        '''
        return _time.tzname[self._isdst(dt)]

    def _isdst(self, dt):
        '''
        Calculate whether the timestamp in question was in daylight savings
        '''
        tt = (dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.weekday(), 0, -1)
        stamp = _time.mktime(tt)
        tt = _time.localtime(stamp)
        return tt.tm_isdst > 0



# -*- coding: utf-8 -*-
'''
Tools for working with paths and files.
'''

from fnmatch import fnmatch
import os
import shutil
import stat


def normalize(*parts):
    '''Join, expand, and normalize a filepath.'''

    path = os.path.expanduser(os.path.expandvars(os.path.join(*parts)))
    return os.path.abspath(path).replace('\\', '/')


def parent(path):
    '''Get a file paths parent directory'''
    return normalize(os.path.dirname(path))


def ensure_path_exists(path, *args):
    '''Like os.makedirs but keeps quiet if path already exists'''

    if os.path.exists(path):
        return

    os.makedirs(path, *args)


def rmtree(path):
    '''Safely remove directory and all of it's contents.'''

    def onerror(func, path, _):
        os.chmod(path, stat.S_IWRITE)
        func(path)

    shutil.rmtree(path, onerror=onerror)


def walk_up(start_dir, depth=20):
    '''Walk up a directory tree'''
    root = start_dir

    for i in range(depth):
        contents = os.listdir(root)
        subdirs, files = [], []
        for f in contents:
            if os.path.isdir(os.path.join(root, f)):
                subdirs.append(f)
            else:
                files.append(f)

        yield root, subdirs, files

        parent = os.path.dirname(root)
        if parent and not parent == root:
            root = parent
        else:
            break


def touch(filepath):
    '''Touch the given filepath'''

    with open(filepath, 'a'):
        os.utime(filepath, None)


def format_size(bytesize):
    '''Human readable size.'''

    value = bytesize
    for unit in ['b', 'kb', 'mb', 'gb', 'tb', 'pb', 'zi']:
        if value < 1024.0:
            return "{:.0f} {}".format(value, unit)
        value /= 1024.0
    return "{:.0f} {}".format(value, unit)


def get_file_count(folder):
    '''Get the number of files in a folder.'''

    count = 0
    for _, _, files in exclusive_walk(folder):
        count += len(files)
    return count


def get_folder_size(folder):
    '''Get the size of a folder in bytes.'''

    size = 0
    for root, _, files in exclusive_walk(folder):
        for file in files:
            file_path = os.path.join(root, file)
            if not os.path.islink(file_path):
                size += os.path.getsize(file_path)
    return size


def is_excluded(value, exclude, exclude_patterns):
    '''Check if a string value matches exclude values and glob patterns'''

    return (
        value in exclude
        or any([fnmatch(value, p) for p in exclude_patterns])
    )


def exclusive_walk(folder, exclude=None, exclude_patterns=None):
    '''Like os.walk but exclude files by value or glob pattern.'''

    exclude = exclude or ['__pycache__', '.git', 'thumbs.db', '.venv', 'venv']
    exclude_patterns = exclude_patterns or ['*.pyc', '*.egg-info']

    for root, subdirs, files in os.walk(folder):
        if is_excluded(os.path.basename(root), exclude, exclude_patterns):
            subdirs[:] = []
            continue

        included = []
        for file in files:
            if is_excluded(file, exclude, exclude_patterns):
                continue
            included.append(file)

        yield root, subdirs, included
